from __future__ import annotations

import abc
import typing

import lime_uow as lu

from lime_etl.domain import (
    job_logger,
    job_status,
    job_test_result,
    value_objects,
)

__all__ = (
    "JobSpec",
    "create_job",
)

UoW = typing.TypeVar("UoW", bound=lu.UnitOfWork, contravariant=True)


class JobSpec(abc.ABC, typing.Generic[UoW]):
    @property
    def dependencies(self) -> typing.Tuple[value_objects.JobName, ...]:
        return tuple()

    @property
    @abc.abstractmethod
    def job_name(self) -> value_objects.JobName:
        raise NotImplementedError

    @property
    def min_seconds_between_refreshes(self) -> value_objects.MinSecondsBetweenRefreshes:
        return value_objects.MinSecondsBetweenRefreshes(0)

    @property
    def min_seconds_between_tests(self) -> value_objects.MinSecondsBetweenTests:
        return value_objects.MinSecondsBetweenTests(0)

    @property
    def max_retries(self) -> value_objects.MaxRetries:
        return value_objects.MaxRetries(0)

    def on_execution_error(self, error_message: str) -> typing.Optional[JobSpec[UoW]]:
        return None

    def on_test_failure(
        self, test_results: typing.FrozenSet[job_test_result.JobTestResult]
    ) -> typing.Optional[JobSpec[UoW]]:
        return None

    @abc.abstractmethod
    def run(
        self,
        uow: UoW,
        logger: job_logger.JobLogger,
    ) -> job_status.JobStatus:
        raise NotImplementedError

    def test(
        self,
        uow: UoW,
        logger: job_logger.JobLogger,
    ) -> typing.List[job_test_result.SimpleJobTestResult]:
        return []

    @property
    def timeout_seconds(self) -> value_objects.TimeoutSeconds:
        return value_objects.TimeoutSeconds(None)

    def __repr__(self) -> str:
        return f"<JobSpec: {self.__class__.__name__}>: {self.job_name.value}"

    def __hash__(self) -> int:
        return hash(self.job_name.value)

    def __eq__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            return (
                self.job_name.value == typing.cast(JobSpec[UoW], other).job_name.value
            )
        else:
            return NotImplemented


class JobSpecImpl(JobSpec[UoW]):
    def __init__(
        # fmt: off
        self,
        *,
        name: str,
        run: typing.Callable[
            [UoW, job_logger.JobLogger],
            job_status.JobStatus
        ],
        test: typing.Optional[
            typing.Callable[
                [UoW, job_logger.JobLogger],
                typing.Optional[typing.List[job_test_result.SimpleJobTestResult]]
            ]
        ] = None,
        dependencies: typing.Optional[typing.Set[str]] = None,
        on_execution_error: typing.Optional[
            typing.Callable[
                [str],
                typing.Optional[JobSpec[UoW]]
            ]
        ] = None,
        on_test_failure: typing.Optional[
            typing.Callable[
                [typing.FrozenSet[job_test_result.JobTestResult]],
                typing.Optional[JobSpec[UoW]]
            ]
        ] = None,
        timeout_seconds: typing.Optional[int] = None,
        max_retries: typing.Optional[int] = None,
        min_seconds_between_refreshes: typing.Optional[int] = None,
        min_seconds_between_tests: typing.Optional[int] = None,
        # fmt: on
    ):
        self._name = value_objects.JobName(name)
        self._run = run
        self._test = test
        self._dependencies = tuple(value_objects.JobName(d) for d in dependencies or [])
        self._on_execution_error = on_execution_error
        self._on_test_failure = on_test_failure
        self._timeout_seconds = value_objects.TimeoutSeconds(timeout_seconds)
        self._max_retries = value_objects.MaxRetries(int(max_retries or 0))
        self._min_seconds_between_refreshes = value_objects.MinSecondsBetweenRefreshes(
            int(min_seconds_between_refreshes or 0)
        )
        self._min_seconds_between_tests = value_objects.MinSecondsBetweenTests(
            int(min_seconds_between_tests or 0)
        )

    @property
    def dependencies(self) -> typing.Tuple[value_objects.JobName, ...]:
        return self._dependencies

    @property
    def job_name(self) -> value_objects.JobName:
        return self._name

    @property
    def min_seconds_between_refreshes(self) -> value_objects.MinSecondsBetweenRefreshes:
        return self._min_seconds_between_refreshes

    @property
    def min_seconds_between_tests(self) -> value_objects.MinSecondsBetweenTests:
        return self._min_seconds_between_tests

    @property
    def max_retries(self) -> value_objects.MaxRetries:
        return self._max_retries

    def on_execution_error(self, error_message: str) -> typing.Optional[JobSpec[UoW]]:
        if self._on_execution_error is None:
            return None
        else:
            return self._on_execution_error(error_message)

    def on_test_failure(
        self, test_results: typing.FrozenSet[job_test_result.JobTestResult]
    ) -> typing.Optional[JobSpec[UoW]]:
        if self._on_test_failure is None:
            return None
        else:
            return self._on_test_failure(test_results)

    def run(
        self,
        uow: UoW,
        logger: job_logger.JobLogger,
    ) -> job_status.JobStatus:
        return self._run(uow, logger)

    def test(
        self,
        uow: UoW,
        logger: job_logger.JobLogger,
    ) -> typing.List[job_test_result.SimpleJobTestResult]:
        if self._test is None:
            return []
        else:
            return self._test(uow, logger) or []

    @property
    def timeout_seconds(self) -> value_objects.TimeoutSeconds:
        return self._timeout_seconds


def create_job(
    # fmt: off
    *,
    name: str,
    run: typing.Callable[[UoW, job_logger.JobLogger], job_status.JobStatus],
    test: typing.Optional[
        typing.Callable[
            [UoW, job_logger.JobLogger],
            typing.Optional[typing.List[job_test_result.SimpleJobTestResult]]
        ]
    ] = None,
    dependencies: typing.Optional[typing.Set[str]] = None,
    on_execution_error: typing.Optional[
        typing.Callable[
            [str],
            typing.Optional[JobSpec[UoW]]
        ]
    ] = None,
    on_test_failure: typing.Optional[
        typing.Callable[
            [typing.FrozenSet[job_test_result.JobTestResult]],
            typing.Optional[JobSpec[UoW]]
        ]
    ] = None,
    timeout_seconds: typing.Optional[int] = None,
    max_retries: typing.Optional[int] = None,
    min_seconds_between_refreshes: typing.Optional[int] = None,
    min_seconds_between_tests: typing.Optional[int] = None,
    # fmt: on
) -> JobSpec[UoW]:
    return typing.cast(
        JobSpec[UoW],
        JobSpecImpl(
            name=name,
            run=run,
            test=test,
            dependencies=dependencies,
            on_execution_error=on_execution_error,
            on_test_failure=on_test_failure,
            timeout_seconds=timeout_seconds,
            max_retries=max_retries,
            min_seconds_between_refreshes=min_seconds_between_refreshes,
            min_seconds_between_tests=min_seconds_between_tests,
        ),
    )



import abc
import datetime

import lime_uow as lu

from lime_etl.domain import value_objects

__all__ = (
    "LocalTimestampAdapter",
    "TimestampAdapter",
)


class TimestampAdapter(lu.Resource[None], abc.ABC):
    @staticmethod
    def key() -> str:
        return TimestampAdapter.__name__

    @abc.abstractmethod
    def now(self) -> value_objects.Timestamp:
        raise NotImplementedError

    def get_elapsed_time(
        self, start_ts: value_objects.Timestamp
    ) -> value_objects.ExecutionMillis:
        end_ts = self.now()
        millis = int((end_ts.value - start_ts.value).total_seconds() * 1000)
        return value_objects.ExecutionMillis(millis)


class LocalTimestampAdapter(TimestampAdapter):
    def now(self) -> value_objects.Timestamp:
        return value_objects.Timestamp(datetime.datetime.now())



import typing

import sqlalchemy as sa
from sqlalchemy import orm

from lime_etl import adapter, domain
from lime_etl.service import admin, batch_runner

__all__ = ("AdminBatch",)


class AdminBatch(domain.BatchSpec[domain.Config, domain.AdminUnitOfWork]):
    def __init__(
        self,
        *,
        config: domain.Config,
        min_seconds_between_runs: domain.MinSecondsBetweenRefreshes = domain.MinSecondsBetweenRefreshes(
            12 * 60 * 60
        ),
        ts_adapter: domain.TimestampAdapter = domain.LocalTimestampAdapter(),
        log_to_console: bool = False,
    ):
        self._config = config
        self._min_seconds_between_runs = min_seconds_between_runs
        self._ts_adapter = ts_adapter
        self._log_to_console = log_to_console

    @property
    def batch_name(self) -> domain.BatchName:
        return domain.BatchName("admin")

    def create_jobs(
        self, uow: domain.AdminUnitOfWork
    ) -> typing.List[domain.JobSpec[domain.AdminUnitOfWork]]:
        return [
            admin.delete_old_logs.DeleteOldLogs(
                days_logs_to_keep=self._config.days_logs_to_keep,
                min_seconds_between_runs=self._min_seconds_between_runs,
            ),
        ]

    def create_uow(self, config: domain.Config) -> domain.AdminUnitOfWork:
        admin_engine = sa.create_engine(self._config.admin_engine_uri.value)
        adapter.admin_metadata.create_all(bind=admin_engine)
        adapter.admin_orm.set_schema(schema=self._config.admin_schema)
        adapter.admin_orm.start_mappers()
        admin_session_factory = orm.sessionmaker(bind=admin_engine)
        return adapter.SqlAlchemyAdminUnitOfWork(
            session_factory=admin_session_factory, ts_adapter=self._ts_adapter
        )

    def run(self) -> domain.BatchStatus:
        return batch_runner.run_batch(
            config=self._config,
            batch=self,
            log_to_console=self._log_to_console,
        )



from sqlalchemy import orm

from lime_etl import domain

__all__ = (
    "SqlAlchemyJobLogger",
    "ConsoleJobLogger",
)


class SqlAlchemyJobLogger(domain.JobLogger):
    def __init__(
        self,
        *,
        batch_id: domain.UniqueId,
        job_id: domain.UniqueId,
        session: orm.Session,
        ts_adapter: domain.TimestampAdapter = domain.LocalTimestampAdapter(),
        log_to_console: bool = False,
    ):
        self._batch_id = batch_id
        self._job_id = job_id
        self._session = session
        self._ts_adapter = ts_adapter
        self._log_to_console = log_to_console

        super().__init__()

    def _log(self, level: domain.LogLevel, message: str) -> None:
        ts = self._ts_adapter.now()
        log_entry = domain.JobLogEntry(
            id=domain.UniqueId.generate(),
            batch_id=self._batch_id,
            job_id=self._job_id,
            log_level=level,
            message=domain.LogMessage(message),
            ts=ts,
        )
        self._session.add(log_entry.to_dto())
        self._session.commit()
        if self._log_to_console:
            print(f"{ts.value.strftime('%H:%M:%S')} [{level.value!s}]: message")

    def error(self, message: str, /) -> None:
        return self._log(
            level=domain.LogLevel.error(),
            message=message,
        )

    def exception(self, /, e: Exception) -> None:
        msg = domain.exceptions.parse_exception(e).text()
        self._log(
            level=domain.LogLevel.error(),
            message=msg,
        )

    def info(self, message: str, /) -> None:
        return self._log(
            level=domain.LogLevel.info(),
            message=message,
        )


class ConsoleJobLogger(domain.JobLogger):
    def __init__(self) -> None:
        super().__init__()

    def error(self, message: str, /) -> None:
        print(f"ERROR: {message}")

    def exception(self, /, e: Exception) -> None:
        print(f"EXCEPTION:\n{domain.exceptions.parse_exception(e).text()}")

    def info(self, message: str, /) -> None:
        print(f"INFO: {message}")



import abc

__all__ = ("JobLogger",)


class JobLogger(abc.ABC):
    @abc.abstractmethod
    def error(self, /, message: str) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def exception(self, /, e: Exception) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def info(self, /, message: str) -> None:
        raise NotImplementedError



import abc

from lime_etl.domain import value_objects

__all__ = ("Config",)


class Config(abc.ABC):
    @property
    @abc.abstractmethod
    def admin_engine_uri(self) -> value_objects.DbUri:
        raise NotImplementedError

    @property
    def admin_schema(self) -> value_objects.SchemaName:
        return value_objects.SchemaName("etl")

    @property
    def days_logs_to_keep(self) -> value_objects.DaysToKeep:
        return value_objects.DaysToKeep(3)



from __future__ import annotations

from lime_uow import sqlalchemy_resources as lsa
from sqlalchemy import orm

__all__ = ("SqlAlchemyAdminSession",)


class SqlAlchemyAdminSession(lsa.SqlAlchemySession):
    def __init__(self, session_factory: orm.sessionmaker):
        super().__init__(session_factory)

    @staticmethod
    def key() -> str:
        return SqlAlchemyAdminSession.__name__



import abc
import functools
import typing

import lime_uow as lu

from lime_etl.domain import cfg, job_spec, value_objects

__all__ = (
    "BatchSpec",
    "create_batch",
)

Cfg = typing.TypeVar("Cfg", bound=cfg.Config)
UoW = typing.TypeVar("UoW", bound=lu.UnitOfWork)


class BatchSpec(abc.ABC, typing.Generic[Cfg, UoW]):
    @functools.cached_property
    def batch_id(self) -> value_objects.UniqueId:
        return value_objects.UniqueId.generate()

    @property
    @abc.abstractmethod
    def batch_name(self) -> value_objects.BatchName:
        raise NotImplementedError

    @abc.abstractmethod
    def create_jobs(self, uow: UoW) -> typing.List[job_spec.JobSpec[UoW]]:
        raise NotImplementedError

    @abc.abstractmethod
    def create_uow(self, config: Cfg) -> UoW:
        raise NotImplementedError

    @property
    def skip_tests(self) -> value_objects.Flag:
        return value_objects.Flag(False)

    @property
    def timeout_seconds(self) -> value_objects.TimeoutSeconds:
        return value_objects.TimeoutSeconds(None)

    def __repr__(self) -> str:
        return f"<BatchSpec: {self.__class__.__name__}>: {self.batch_name.value}"

    def __hash__(self) -> int:
        return hash(self.batch_name.value)

    def __eq__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            return (
                self.batch_name.value
                == typing.cast(
                    BatchSpec[typing.Any, typing.Any], other
                ).batch_name.value
            )
        else:
            return NotImplemented


class BatchSpecImpl(BatchSpec[Cfg, UoW]):
    def __init__(
        # fmt: off
        self,
        *,
        name: str,
        create_uow: typing.Callable[[Cfg], UoW],
        jobs: typing.List[job_spec.JobSpec[UoW]],
        runtime_jobs: typing.Optional[
            typing.Callable[[UoW], typing.List[job_spec.JobSpec[UoW]]]
        ],
        skip_tests: bool,
        timeout_seconds: typing.Optional[int],
        # fmt: on
    ):
        self._batch_name = value_objects.BatchName(name)
        self._jobs = jobs
        self._runtime_jobs = runtime_jobs
        self._create_uow = create_uow
        self._skip_tests = value_objects.Flag(skip_tests)
        self._timeout_seconds = value_objects.TimeoutSeconds(timeout_seconds)

    @functools.cached_property
    def batch_id(self) -> value_objects.UniqueId:
        return value_objects.UniqueId.generate()

    @property
    def batch_name(self) -> value_objects.BatchName:
        return self._batch_name

    def create_jobs(self, uow: UoW) -> typing.List[job_spec.JobSpec[UoW]]:
        if self._runtime_jobs:
            return self._runtime_jobs(uow) + self._jobs
        else:
            return self._jobs

    def create_uow(self, config: Cfg) -> UoW:
        return self._create_uow(config)

    @property
    def skip_tests(self) -> value_objects.Flag:
        return value_objects.Flag(False)

    @property
    def timeout_seconds(self) -> value_objects.TimeoutSeconds:
        return value_objects.TimeoutSeconds(None)


def create_batch(
    # fmt: off
    *,
    name: str,
    jobs: typing.List[job_spec.JobSpec[UoW]],
    create_uow: typing.Callable[[Cfg], UoW],
    runtime_jobs: typing.Optional[
        typing.Callable[
            [UoW],
            typing.List[job_spec.JobSpec[UoW]]
        ]
    ] = None,
    skip_tests: bool = False,
    timeout_seconds: typing.Optional[int] = None,
    # fmt: on
) -> BatchSpec[Cfg, UoW]:
    return typing.cast(
        BatchSpec[Cfg, UoW],
        BatchSpecImpl(
            name=name,
            jobs=jobs,
            runtime_jobs=runtime_jobs,
            create_uow=create_uow,
            skip_tests=skip_tests,
            timeout_seconds=timeout_seconds,
        ),
    )
